@@ -15,7 +15,7 @@ use {
             ext_debug_utils::EXT_DEBUG_UTILS_EXTENSION_NAME,
             khr_surface::KHR_SURFACE_EXTENSION_NAME,
         },
-        utils::loading::{DefaultEntryLoader, EntryLoaderError},
+        utils::loading::{EntryLoader, EntryLoaderError},
         vk1_0, InstanceLoader, LoaderError,
     },
     once_cell::sync::OnceCell,
@@ -52,16 +52,22 @@ use erupt::extensions::khr_win32_surface::{
     Win32SurfaceCreateInfoKHR, KHR_WIN32_SURFACE_EXTENSION_NAME,
 };
 
-#[cfg(any(target_os = "ios", target_os = "macos"))]
-use erupt::extensions::ext_metal_surface::{
-    MetalSurfaceCreateInfoEXT, EXT_METAL_SURFACE_EXTENSION_NAME,
+#[cfg(target_os = "ios")]
+use erupt::{
+    extensions::mvk_ios_surface::MVK_IOS_SURFACE_EXTENSION_NAME, vk::IOSSurfaceCreateInfoMVKBuilder,
+};
+
+#[cfg(target_os = "macos")]
+use erupt::{
+    extensions::mvk_macos_surface::MVK_MACOS_SURFACE_EXTENSION_NAME,
+    vk::MacOSSurfaceCreateInfoMVKBuilder,
 };
 
 /// Root object of the erupt graphics system.
 pub struct Graphics {
     pub(crate) instance: InstanceLoader,
     pub(crate) version: u32,
-    _entry: DefaultEntryLoader,
+    _entry: EntryLoader,
 }
 
 static GLOBAL_GRAPHICS: OnceCell<Graphics> = OnceCell::new();
@@ -110,7 +116,7 @@ impl Graphics {
     fn new() -> Result<Self, InitError> {
         tracing::trace!("Init erupt graphisc implementation");
 
-        let entry = DefaultEntryLoader::new()?;
+        let entry = EntryLoader::new()?;
 
         let version = entry.instance_version();
 
@@ -224,27 +230,34 @@ impl Graphics {
                 push_ext(KHR_WIN32_SURFACE_EXTENSION_NAME);
             }
 
-            #[cfg(any(target_os = "ios", target_os = "macos"))]
+            #[cfg(target_os = "ios")]
             {
-                push_ext(EXT_METAL_SURFACE_EXTENSION_NAME);
+                push_ext(MVK_IOS_SURFACE_EXTENSION_NAME);
+            }
+
+            #[cfg(target_os = "macos")]
+            {
+                push_ext(MVK_MACOS_SURFACE_EXTENSION_NAME);
             }
         }
 
-        let result = InstanceLoader::new(
-            &entry,
-            &vk1_0::InstanceCreateInfoBuilder::new()
-                .application_info(
-                    &vk1_0::ApplicationInfoBuilder::new()
-                        .engine_name(CStr::from_bytes_with_nul(b"Illume\0").unwrap())
-                        .engine_version(1)
-                        .application_name(CStr::from_bytes_with_nul(b"IllumeApp\0").unwrap())
-                        .application_version(1)
-                        .api_version(version),
-                )
-                .enabled_layer_names(&enable_layers)
-                .enabled_extension_names(&enable_exts),
-            None,
-        );
+        let result = unsafe {
+            InstanceLoader::new(
+                &entry,
+                &vk1_0::InstanceCreateInfoBuilder::new()
+                    .application_info(
+                        &vk1_0::ApplicationInfoBuilder::new()
+                            .engine_name(CStr::from_bytes_with_nul(b"Illume\0").unwrap())
+                            .engine_version(1)
+                            .application_name(CStr::from_bytes_with_nul(b"IllumeApp\0").unwrap())
+                            .application_version(1)
+                            .api_version(version),
+                    )
+                    .enabled_layer_names(&enable_layers)
+                    .enabled_extension_names(&enable_exts),
+                None,
+            )
+        };
 
         let instance = match result {
             Err(LoaderError::SymbolNotAvailable) => {
@@ -263,7 +276,6 @@ impl Graphics {
                         .flags(DebugReportFlagsEXT::all())
                         .pfn_callback(Some(debug_report_callback)),
                     None,
-                    None,
                 )
             }
             .result()?;
@@ -281,7 +293,7 @@ impl Graphics {
     }
 
     pub fn name(&self) -> &str {
-        "Erupt"
+        "Vulkan"
     }
 
     #[tracing::instrument]
@@ -328,21 +340,88 @@ impl Graphics {
                     });
                 }
 
-                let result =
-                    self.instance
-                        .create_android_surface_khr(&AndroidSurfaceCreateInfoKHR {
-                            window: handle.a_native_window,
-                            ..AndroidSurfaceCreateInfoKHR::default()
-                        });
+                let result = self.instance.create_android_surface_khr(
+                    &AndroidSurfaceCreateInfoKHR {
+                        window: handle.a_native_window,
+                        ..AndroidSurfaceCreateInfoKHR::default()
+                    },
+                    None,
+                );
 
                 todo!()
             }
 
             #[cfg(target_os = "ios")]
-            RawWindowHandle::IOS(handle) => todo!(),
+            RawWindowHandle::IOS(handle) => {
+                todo!()
+            }
 
             #[cfg(target_os = "macos")]
-            RawWindowHandle::MacOS(handle) => todo!(),
+            RawWindowHandle::MacOS(handle) => {
+                use core_graphics_types::{base::CGFloat, geometry::CGRect};
+                use objc::{
+                    class, msg_send,
+                    runtime::{Object, BOOL, YES},
+                    sel, sel_impl,
+                };
+
+                if !self.instance.enabled().mvk_macos_surface {
+                    return Err(CreateSurfaceError::UnsupportedWindow {
+                        window: RawWindowHandleKind::Windows,
+                        source: Some(Box::new(RequiredExtensionIsNotAvailable {
+                            extension: "VK_MVK_macos_surface",
+                        })),
+                    });
+                }
+
+                let class = class!(CAMetalLayer);
+
+                unsafe {
+                    let view = handle.ns_view as *mut Object;
+                    if view.is_null() {
+                        panic!("window does not have a valid contentView");
+                    }
+
+                    let existing: *mut Object = msg_send![view, layer];
+                    let use_current = if existing.is_null() {
+                        false
+                    } else {
+                        let result: BOOL = msg_send![existing, isKindOfClass: class];
+                        result == YES
+                    };
+
+                    if !use_current {
+                        let layer: mtl::MetalLayer = msg_send![class, new];
+                        let () = msg_send![view, setLayer: layer.as_ref()];
+                        let () = msg_send![view, setWantsLayer: YES];
+                        let bounds: CGRect = msg_send![view, bounds];
+                        let () = msg_send![layer.as_ref(), setBounds: bounds];
+
+                        let window: *mut Object = msg_send![view, window];
+                        if !window.is_null() {
+                            let scale_factor: CGFloat = msg_send![window, backingScaleFactor];
+                            let () = msg_send![layer, setContentsScale: scale_factor];
+                        }
+                    };
+                }
+
+                let result = unsafe {
+                    self.instance.create_mac_os_surface_mvk(
+                        &MacOSSurfaceCreateInfoMVKBuilder::new().view(handle.ns_view),
+                        None,
+                    )
+                }
+                .result();
+
+                match result {
+                    Ok(surface) => surface,
+                    Err(vk1_0::Result::ERROR_OUT_OF_HOST_MEMORY) => out_of_host_memory(),
+                    Err(vk1_0::Result::ERROR_OUT_OF_DEVICE_MEMORY) => {
+                        return Err(OutOfMemory.into())
+                    }
+                    Err(err) => unexpected_result(err),
+                }
+            }
 
             #[cfg(any(
                 target_os = "linux",
@@ -368,7 +447,6 @@ impl Graphics {
                             .surface(handle.surface)
                             .display(handle.display),
                         None,
-                        None,
                     )
                 }
                 .result();
@@ -398,7 +476,7 @@ impl Graphics {
                     let mut info = Win32SurfaceCreateInfoKHR::default();
                     info.hinstance = handle.hinstance;
                     info.hwnd = handle.hwnd;
-                    self.instance.create_win32_surface_khr(&info, None, None)
+                    self.instance.create_win32_surface_khr(&info, None)
                 }
                 .result()
                 .map_err(|err| match err {
@@ -432,7 +510,6 @@ impl Graphics {
                             .window(handle.window)
                             .connection(handle.connection),
                         None,
-                        None,
                     )
                 }
                 .result();
@@ -472,7 +549,6 @@ impl Graphics {
                             ..XlibSurfaceCreateInfoKHR::default()
                         },
                         None,
-                        None,
                     )
                 }
                 .result()
